from .__version__ import (__version__)  # noqa: F401
from .cnd_scaffold import CndScaffold  # noqa: F401
from .scaffold import Scaffold  # noqa: F401
from .file import File  # noqa: F401
from .structure import Structure  # noqa: F401



import yamale
import yaml
import json
import jinja2


class CndScaffold:
    step = {'init', 'build', 'runtime'}
    allowed_engine = ['default', 'jinja2']

    def __init__(self, source, target, data_to_replace, cnd_io, print, engine='default', cnd_io_target=None):
        self._print = print
        self._source = source
        self._target = target
        self._cnd_io = cnd_io
        self.engine = engine
        if cnd_io_target is None:
            self._cnd_io_target = cnd_io
        else:
            self._cnd_io_target = cnd_io_target
        self._data_to_replace = data_to_replace

    def _load_schema(self):
        self.schema = self._cnd_io.pull_file(self._source['project_id'], 'schema.yml', branch=self._source['branch'])
        if self.schema is False:
            self._print.info_e(f"Cannot found file for schema : {self.schema}")
            raise NameError('Schema unreadable')
        return self.schema

    def _load_structure(self):
        self.structure = self._cnd_io.pull_file(self._source['project_id'], f'{self._source["definition"]}.yml', branch=self._source['branch'])
        if self.structure is False:
            self._print.info_e(f"Cannot found file for structure : {self.structure}")
            raise NameError('Structure unreadable')
        return self.structure

    def _validate_structure(self):
        self._load_schema()
        self._load_structure()
        schema = yamale.make_schema(content=self.schema)
        data = yamale.make_data(content=self.structure)
        yamale.validate(schema, data)
        content_yaml = yaml.safe_load(self.structure)
        self.yaml_structure = {'source_folder': content_yaml['source_folder']}
        for item in content_yaml['files']:
            for step in item['step']:
                if step not in self.yaml_structure:
                    self.yaml_structure[step] = []
                if 'step' in item:
                    del item['step']
                self.yaml_structure[step].append(item)
        return self.yaml_structure

    def _pull_file(self, project_id, file_name, branch):
        return self._cnd_io.pull_file(project_id, file_name, branch=branch)

    def _commit_file(self, project_id, file_name, content, branch):
        return self._cnd_io_target.commit_file(project_id, file_name, content, branch)

    def _replace_default(self, item, target, content, sort=True):
        files = []
        repeat = {'values': [''], 'key': ''}
        if 'repeat' in item:
            repeat = {'key': item['repeat'], 'values': self._data_to_replace[item['repeat']]}
        for value in repeat['values']:
            if  value == '':
                global_vars = self._data_to_replace
            else:
                global_vars = {repeat['key']: value}
            for item in self._data_to_replace:
                if isinstance(self._data_to_replace[item], list) or isinstance(self._data_to_replace[item], dict):
                    continue
                global_vars[item] = self._data_to_replace[item]
            file_name = self._replace_content(target, global_vars)
            file_content = self._replace_content(content, global_vars)
            if file_name.split('.')[-1] == 'yml':
                try:
                    file_content = yaml.safe_load(file_content)
                    file_content = yaml.dump(file_content, sort_keys=sort)
                except yaml.scanner.ScannerError:
                    self._print.info_e('Yaml file is not valide')
                    self._print.info_e(file_content)
                    raise yaml.scanner.ScannerError('A very specific bad thing happened')
            files.append({'file_name': file_name.strip(), 'file_content': file_content.strip()})
        return files

    def _load_file(self, item):
        sort = item["sort"] if 'sort' in item else True
        content = self._pull_file(self._source['project_id'], f'{self.yaml_structure["source_folder"]}/{item["name"]}', self._source['branch'])
        if content is False:
            self._print.info_e(f'Cannot found file for : {self._source["project_id"]} - {self.yaml_structure["source_folder"]}/{item["name"]}')
            raise NameError(f'File unreadable {self.yaml_structure["source_folder"]}/{item["name"]}')
        target = item['name']
        if 'target' in item:
            target = item['target']
        files = self._replace_default(item, target, content, sort=sort)
        for file in files:
            self._commit_file(self._target['project_id'],
                f"{self._target['folder']}/{file['file_name']}",
                file["file_content"],
                self._target['branch']
            )
        return files

    def _load_files(self, step):
        self.file_to_upload = []
        if step not in self.yaml_structure:
            return 0
        total = 0
        for item in self.yaml_structure[step]:
            total += len(self._load_file(item))
        return total

    def replace_multi_type(self, content, search, value):
        str_value = value if isinstance(value, list) is False else ",".join(value)
        content = content.replace(f"%{search}%", str_value)
        return content

    def _default_replace_engine(self, content, local_var={}):
        for item in local_var:
            if isinstance(local_var[item], dict) is False:
                content = self.replace_multi_type(content, item, local_var[item])
                continue
            for key in local_var[item]:
                str_to_replace = f'{item}.{key}'
                content = self.replace_multi_type(content, str_to_replace, local_var[item][key])
        for item in self._data_to_replace:
            if isinstance(self._data_to_replace[item], dict) is True or isinstance(self._data_to_replace[item], list) is True:
                content = content.replace(f"%{item}%", json.dumps(self._data_to_replace[item]))
            else:
                content = content.replace(f"%{item}%", self._data_to_replace[item])
        return content

    def _replace_content(self, content, local_var={}):
        if self.engine == 'default':
            return self._default_replace_engine(content, local_var=local_var)
        else:
            environment = jinja2.Environment()
            template = environment.from_string(content)
            return template.render(local_var)

    def _push_files(self):
        result = self._cnd_io_target.push_files(self._target['project_id'], commit_message="Pushing from CndScaffold", branch=self._target["branch"])
        if result is False:
            self._print.info_e(f'Cannot push content')
            self._print.trace_e(f'Content: {yaml.dump(self._cnd_io_target._files)}')
            raise NameError(f'Content cannot be pushed')
        return True

    def _apply_step(self, step):
        self._validate_structure()
        self._load_files(step)
        return self._push_files()

    def init(self):
        return self._apply_step('init')

    def build(self):
        return self._apply_step('build')

    def runtime(self):
        return self._apply_step('runtime')



import cndprint  # noqa: E402
import cnd_io  # noqa: E402
import yaml


level = "Trace"
silent_mode = True
_print = cndprint.CndPrint(level=level, silent_mode=silent_mode)
_provider = cnd_io.CndProviderLocalfile(creds={}, print=_print)
_cnd_io = cnd_io.CndIO(_provider, print=_print)
_cnd_io_target = cnd_io.CndIO(_provider, print=_print)


def read_file(filename):
    return open(filename).read()

def read_yaml_file(filename):
    file = read_file(filename)
    return yaml.safe_load(file)

def compare_text(result, expected):
    text1 = result.splitlines()
    text2 = expected.splitlines()
    for line in difflib.unified_diff(text1, text2, fromfile="result", tofile="expected, lineterm='"):
        _print.info_d(line)


from yamale.yamale_error import YamaleError  # noqa: E402
from mockito import when, unstub  # noqa: E402
from expects import *  # noqa: F403, E402
from mamba import description, context, it  # noqa: E402
import yaml  # noqa: E402
import src.cnd_scaffold as scaffold
import tests.vars as vars


project = {}
definition = {}

with description("Structure with provider") as self:
    with context("__init__"):
        with it("should init the print in the class"):
            unstub()
            self.tested_instance = scaffold.structure.Structure(project, definition, vars._print, provider="A")
            expect(self.tested_instance._provider).to(equal("A"))

with description("Structure") as self:
    with before.each:
        unstub()
        self.tested_instance = scaffold.structure.Structure(project, definition, vars._print)

    with context("__init__"):
        with it("should init the print in the class"):
            expect(self.tested_instance._print).to(equal(vars._print))

    with context("structure getter"):
        with it("should return structure value"):
            when(self.tested_instance._cnd_io).pull_file(...).thenReturn('A')
            expect(self.tested_instance.structure).to(equal('A'))

        with it("should generate an error"):
            when(self.tested_instance._cnd_io).pull_file(...).thenReturn(False)
            expect(lambda: self.tested_instance.structure).to(raise_error(NameError))

    with context("schema getter"):
        with it("should return schema value"):
            when(self.tested_instance._cnd_io).pull_file(...).thenReturn('A')
            expect(self.tested_instance.schema).to(equal('A'))

        with it("should generate an error"):
            when(self.tested_instance._cnd_io).pull_file(...).thenReturn(False)
            expect(lambda: self.tested_instance.schema).to(raise_error(NameError))

    with context("files getter"):
        with it("should return schema value"):
            self.tested_instance._structure = vars.read_file('tests/data/structure/structure.yml')
            expected = vars.read_yaml_file('tests/data/structure/expected.yml')
            when(self.tested_instance._cnd_io).pull_file(...).thenReturn("A").thenReturn("B").thenReturn("C")
            expect(self.tested_instance.files).to(equal(expected))

        with it("should return cached value"):
            self.tested_instance._files = 'ABA'
            expect(self.tested_instance.files).to(equal('ABA'))

    with context("_validate"):
        with it("should return true"):
            self.tested_instance._schema = vars.read_file('tests/data/structure/schema.yml')
            self.tested_instance._structure = vars.read_file('tests/data/structure/structure.yml')
            result = self.tested_instance._validate()
            expect(result).to(equal(True))

        with it("should raise an error"):
            self.tested_instance._schema = vars.read_file('tests/data/structure/schema.yml')
            self.tested_instance._structure = vars.read_file('tests/data/structure/bad_structure.yml')
            expect(lambda: self.tested_instance._validate()).to(raise_error(NameError))   

    with context("get"):
        with it("should return file"):
            files = vars.read_yaml_file('tests/data/structure/expected.yml')
            self.tested_instance._files = files
            when(self.tested_instance)._validate(...).thenReturn(True)
            expect(self.tested_instance.get()).to(equal(files))        


from yamale.yamale_error import YamaleError  # noqa: E402
from mockito import when, unstub  # noqa: E402
from expects import *  # noqa: F403, E402
from mamba import description, context, it  # noqa: E402
import cndprint  # noqa: E402
import cnd_io  # noqa: E402
import yaml  # noqa: E402
import src.cnd_scaffold.cnd_scaffold as cnd_scaffold


level = "Trace"
_print = cndprint.CndPrint(level=level, silent_mode=True)
provider = cnd_io.CndProviderLocalfile(creds={}, print=_print)
_cnd_io = cnd_io.CndIO(provider, print=_print)
_cnd_io_target = cnd_io.CndIO(provider, print=_print)
source = {
    'project_id': 'tests/demo',
    'definition': 'org-demo1_product1',
    'branch': 'main',
}
target = {
    'project_id': 'tests/demo-result',
    'folder': 'home',
    'branch': 'main',
}
data_to_replace = {
    'env': ['alpha', 'gamma'],
    'client': [{'name': 'A', 'token': 'B'}, {'name': 'C', 'token': 'D'}],
    'app': 'beta',
    'abc': 'def',
    'yaml': {
        'abc': 'def',
        'ghi': ['jkl', 'lmo']
    }
}
with description("CndScaffold") as self:
    with before.each:
        unstub()
        self.cnd_scaffold = cnd_scaffold.CndScaffold(source, target, data_to_replace, _cnd_io, print=_print)

    with context("__init__"):
        with it("should init the print in the class"):
            expect(self.cnd_scaffold._print).to(equal(_print))

with description("CndScaffold") as self:
    with before.each:
        unstub()
        self.cnd_scaffold = cnd_scaffold.CndScaffold(source, target, data_to_replace, _cnd_io, print=_print, cnd_io_target=_cnd_io_target)

    with context("__init__"):
        with it("should init the print in the class"):
            expect(self.cnd_scaffold._print).to(equal(_print))

    with context("_pull_file"):
        with it("should init the print in the class"):
            when(self.cnd_scaffold._cnd_io).pull_file(...).thenReturn(True)
            expect(self.cnd_scaffold._pull_file('project_id', 'filename', 'branch')).to(equal(True))

    with context("_commit_file"):
        with it("should init the print in the class"):
            when(self.cnd_scaffold._cnd_io_target).commit_file(...).thenReturn(True)
            expect(self.cnd_scaffold._commit_file('project_id', 'filename', 'content', 'branch')).to(equal(True))

    with context("_load_schema"):
        with it("should return the scheme if scheme is ok"):
            content = provider.pull_file('tests/demo', 'schema.yml')
            result = self.cnd_scaffold._load_schema()
            expect(result).to(equal(content))

        with it("should raise an error if scheme is not ok"):
            when(self.cnd_scaffold._cnd_io).pull_file(...).thenReturn(False)
            expect(lambda: self.cnd_scaffold._load_schema()).to(raise_error(NameError))

    with context("_load_structure"):
        with it("should return the structure if structure is ok"):
            content = provider.pull_file('tests/demo', 'org-demo1_product1.yml')
            result = self.cnd_scaffold._load_structure()
            expect(result).to(equal(content))

        with it("should raise an error if structure is not ok"):
            when(self.cnd_scaffold._cnd_io).pull_file(...).thenReturn(False)
            expect(lambda: self.cnd_scaffold._load_structure()).to(raise_error(NameError))

    with context("_validate_structure"):
        with it("should return the structure if structure is ok"):
            content = provider.pull_file('tests/data', 'yaml_structure.yml')
            result = self.cnd_scaffold._validate_structure()
            expect(result).to(equal(yaml.safe_load(content)))

    with context("_replace_default"):
        with it("should return file"):
            self.cnd_scaffold._target = target
            self.cnd_scaffold._data_to_replace = {'name': 'abc'}
            result = self.cnd_scaffold._replace_default({}, '%name%.yml', 'abc: %name%\n\np: a')
            expect(result).to(equal([{'file_name': 'abc.yml', 'file_content': 'abc: abc\np: a'}]))

        with it("should return file with repeat"):
            self.cnd_scaffold._target = target
            self.cnd_scaffold._data_to_replace = {'name': 'abc', 'mykey': ['a']}
            result = self.cnd_scaffold._replace_default({'repeat': 'mykey'}, '%name%.yml', 'abc: %name%\n\np: a')
            expect(result).to(equal([{'file_name': 'abc.yml', 'file_content': 'abc: abc\np: a'}]))

        with it("should return good content with jinja engine"):
            self.cnd_scaffold.engine = 'jinja2'
            self.cnd_scaffold._target = target
            self.cnd_scaffold._data_to_replace = {'name': 'abc'}
            result = self.cnd_scaffold._replace_default({}, '{{ name }}.yml', 'abc: {{ name }}\n\np: a')
            expect(result).to(equal([{'file_name': 'abc.yml', 'file_content': 'abc: abc\np: a'}]))

        # with it("should return good content with jinja engine and loop"):
        #     self.cnd_scaffold.engine = 'jinja2'
        #     self.cnd_scaffold._target = target
        #     self.cnd_scaffold._data_to_replace = {'clients': {'A': {'lines': {'C': 1, 'A': 2}}}, 'lines': {'C': 1, 'A': 2}}
        #     result = self.cnd_scaffold._replace_default({'repeat': 'clients'}, '{{ clients }}.yml', '{% for line in lines %}{{ line }}{% endfor %}')
        #     expect(result).to(equal([{'file_name': 'abc.yml', 'file_content': 'abc: abc\np: a'}]))

        with it("should throw error if invalid yaml"):
            self.cnd_scaffold.engine = 'jinja2'
            self.cnd_scaffold._target = target
            self.cnd_scaffold._data_to_replace = {'name': 'abc'}
            expect(lambda: self.cnd_scaffold._replace_default({}, '{{ name }}.yml', 'abc: {{ name }}\n\np: b: a')).to(raise_error(yaml.scanner.ScannerError))

    with context("_validate_structure"):
        with it("should raise an error if structure is not valid"):
            when(self.cnd_scaffold)._load_structure(...).thenReturn()
            self.cnd_scaffold.structure = '{a: b}'
            expect(lambda: self.cnd_scaffold._validate_structure()).to(raise_error(YamaleError))

    with context("_load_file"):
        with before.each:
            self.cnd_scaffold.yaml_structure = yaml.safe_load(provider.pull_file('tests/data', 'yaml_structure.yml'))
            when(self.cnd_scaffold)._pull_file(...).thenReturn('')

        with it("should load and prepare file"):
            self.cnd_scaffold.engine = 'jinja2'
            when(self.cnd_scaffold)._commit_file(...).thenReturn(2)
            when(self.cnd_scaffold)._replace_default(...).thenReturn([{'file_name': 'file_name', 'file_content': 'file_content'}])
            result = self.cnd_scaffold._load_file({"name": 'abc', 'target': 'def'})
            expect(len(result)).to(equal(1))

        with it("should load and prepare file"):
            self.cnd_scaffold.engine = 'jinja2'
            when(self.cnd_scaffold)._commit_file(...).thenReturn(4)
            when(self.cnd_scaffold)._replace_default(...).thenReturn([{'file_name': 'file_name', 'file_content': 'file_content'}, {'file_name': 'file_name', 'file_content': 'file_content'}])
            result = self.cnd_scaffold._load_file({"name": 'abc'})
            expect(len(result)).to(equal(2))

        with it("should load and prepare file"):
            when(self.cnd_scaffold)._commit_file(...).thenReturn(3)
            when(self.cnd_scaffold)._replace_default(...).thenReturn([{'file_name': 'file_name', 'file_content': 'file_content'}])
            result = self.cnd_scaffold._load_file({"name": 'abc'})
            expect(len(result)).to(equal(1))

    with context("_load_file (loop)"):
        with before.each:
            self.cnd_scaffold.yaml_structure = yaml.safe_load(provider.pull_file('tests/data', 'yaml_structure.yml'))
            when(self.cnd_scaffold)._pull_file(...).thenReturn(provider.pull_file('tests/demo/org-demo1', 'product1/loop/loop.yml'))
            self.expected_result = [
                {'file_name': 'ad-A/B.yml', 'file_content': 'name: B'},
                {'file_name': 'ad-C/D.yml', 'file_content': 'name: D'}
            ]

        with it("should load and prepare file"):
            my_file = self.cnd_scaffold.yaml_structure['init'][-2]
            result = self.cnd_scaffold._load_file(my_file)
            expect(result).to(equal(self.expected_result))

        with it("should load and prepare file"):
            when(self.cnd_scaffold)._pull_file(...).thenReturn(provider.pull_file('tests/demo/org-demo1', 'product1/loop-jinja/loop.yml'))
            my_file = self.cnd_scaffold.yaml_structure['init'][-1]
            self.cnd_scaffold.engine = 'jinja2'
            result = self.cnd_scaffold._load_file(my_file)
            expect(result).to(equal(self.expected_result))

    with context("_load_files"):
        with before.each:
            self.cnd_scaffold.yaml_structure = yaml.safe_load(provider.pull_file('tests/data', 'yaml_structure.yml'))
            when(self.cnd_scaffold)._pull_file(...).thenReturn('')
            when(self.cnd_scaffold)._commit_file(...).thenReturn(5)

        with it("should load and prepare file"):
            when(self.cnd_scaffold)._load_file(...).thenReturn([{}])
            result = self.cnd_scaffold._load_files('init')
            expect(result).to(equal(7))

        with it("should raise an error if file cannot be load"):
            when(self.cnd_scaffold)._pull_file(...).thenReturn(False)
            expect(lambda: self.cnd_scaffold._load_files('init')).to(raise_error(NameError))


        with it("should skip and return 0 if nothing to push for this step"):
            expect(self.cnd_scaffold._load_files('toto')).to(equal(0))

    with context("_replace_content for jinja engine"):
        with before.each:
            self.cnd_scaffold.engine = 'jinja2'

        with it("should replace data"):
            template = "A{{ varname }}C{{ othervar}}E"
            result = self.cnd_scaffold._replace_content(template, {'varname': 'B', 'othervar': 'D'})
            expect(result).to(equal('ABCDE'))

    with context("_replace_content"):
        with before.each:
            self.cnd_scaffold.engine = 'default'

        with it("should replace when data is an array of dict"):
            result = self.cnd_scaffold._replace_content("%client.name%", {'client': data_to_replace['client'][0]})
            expect(result).to(equal('A'))

        with it("should replace"):
            result = self.cnd_scaffold._replace_content("%abc%-%app%", {})
            expect(result).to(equal(f'{data_to_replace["abc"]}-{data_to_replace["app"]}'))

        with it("should replace"):
            result = self.cnd_scaffold._replace_content("%abc%-%app%", {})
            expect(result).to(equal(f'{data_to_replace["abc"]}-{data_to_replace["app"]}'))

        with it("should replace"):
            self.cnd_scaffold._data_to_replace = {
                'app': 'beta',
            }
            result = self.cnd_scaffold._replace_content("%abc%-%app%", {})
            expect(result).to(equal(f'%abc%-{data_to_replace["app"]}'))

        with it("should replace with the local_var first"):
            result = self.cnd_scaffold._replace_content("%abc%-", {'abc': 'me-first'})
            expect(result).to(equal(f'me-first-'))

        with it("should replace with the local_var first"):
            result = self.cnd_scaffold._replace_content("a: %env%", {})
            expect(result).to(equal('a: ["alpha", "gamma"]'))

    with context("_push_files"):
        with it("should push files"):
            when(self.cnd_scaffold._cnd_io).push_files(...).thenReturn(True)
            result = self.cnd_scaffold._push_files()
            expect(result).to(equal(True))

        with it("should raise an error if push failed"):
            when(self.cnd_scaffold._cnd_io_target).push_files(...).thenReturn(False)
            expect(lambda: self.cnd_scaffold._push_files()).to(raise_error(NameError))

    with context("_apply_step"):
        with before.each:
            when(self.cnd_scaffold)._validate_structure(...).thenReturn()
            when(self.cnd_scaffold)._load_files(...).thenReturn()

        with it("should return True if everything work"):
            when(self.cnd_scaffold)._push_files(...).thenReturn(True)
            result = self.cnd_scaffold._apply_step('')
            expect(result).to(equal(True))

        with it("should raise an error if push failed"):
            when(self.cnd_scaffold)._push_files(...).thenRaise(NameError)
            expect(lambda: self.cnd_scaffold._apply_step('')).to(raise_error(NameError))

    with context("init"):
        with it("should return True if everything work"):
            when(self.cnd_scaffold)._apply_step(...).thenReturn(True)
            result = self.cnd_scaffold.init()
            expect(result).to(equal(True))

        with it("should raise an error if push failed"):
            when(self.cnd_scaffold)._apply_step(...).thenRaise(NameError)
            expect(lambda: self.cnd_scaffold.init()).to(raise_error(NameError))

    with context("build"):
        with it("should return True if everything work"):
            when(self.cnd_scaffold)._apply_step(...).thenReturn(True)
            result = self.cnd_scaffold.build()
            expect(result).to(equal(True))

        with it("should raise an error if push failed"):
            when(self.cnd_scaffold)._apply_step(...).thenRaise(NameError)
            expect(lambda: self.cnd_scaffold.build()).to(raise_error(NameError))

    with context("runtime"):
        with it("should return True if everything work"):
            when(self.cnd_scaffold)._apply_step(...).thenReturn(True)
            result = self.cnd_scaffold.runtime()
            expect(result).to(equal(True))

        with it("should raise an error if push failed"):
            when(self.cnd_scaffold)._apply_step(...).thenRaise(NameError)
            expect(lambda: self.cnd_scaffold.runtime()).to(raise_error(NameError))

    with context("_default_replace_engine"):
        with before.each:
            f = open("tests/demo/org-demo1/product1/loop/loop.yml", "r")
            self.client_file = f.read()

        with it("should replace string"):
            result = self.cnd_scaffold._default_replace_engine(self.client_file, {"client.token": 'Name'})
            expect(result).to(equal('name: Name\n'))

        with it("should replace string"):
            result = self.cnd_scaffold._default_replace_engine(self.client_file, {"client.token": ['Na', 'me']})
            expect(result).to(equal('name: Na,me\n'))

        with it("should replace string"):
            result = self.cnd_scaffold._default_replace_engine(self.client_file, {"client": {'token': 'Name'}})
            expect(result).to(equal('name: Name\n'))

        with it("should replace string"):
            result = self.cnd_scaffold._default_replace_engine(self.client_file, {"client": {'token': ['Na', 'me']}})
            expect(result).to(equal('name: Na,me\n'))
